""" Sound Jubilee app. Demonstrates sound and music. """

import os
from jubilee import App, Mode
from jubilee.controls import Button

class SoundApp(App):
	""" Sound app. """

	def init(self):
		self.add_mode(SoundMode)

class SoundMode(Mode):
	""" Sound mode. """

	def init(self):

		self.name = 'Sound'

		# load music filenames
		music_folder = os.path.join(self.app.base_path, 'music')
		self.music = list(f for f in os.listdir(music_folder) if os.path.splitext(f)[1].lower() == '.mp3')

		self.music_index = 0
		buttons = [['Next', self.next_music], ['Play', self.play_music],
			['Stop', self.stop_music]]
		for i, (text, handler) in enumerate(buttons):
			self.add_control(Button(text, 100+50*i, 50, 40, 40, click=handler))

		self.sound_index = 0
		buttons = [['Next', self.next_sound], ['Play', self.play_sound]]
		for i, (text, handler) in enumerate(buttons):
			self.add_control(Button(text, 100+50*i, 130, 40, 40, click=handler))

	def next_music(self):
		""" Click handler for Next Music button. """

		self.music_index = (self.music_index + 1) % len(self.music)

	def play_music(self):
		""" Click handler for Play Music button. """

		self.app.play_music(self.music[self.music_index])

	def stop_music(self):
		""" Click handler for Stop Music button. """

		self.app.stop_music()

	def next_sound(self):
		""" Click handler for Next Sound button. """

		self.sound_index = (self.sound_index + 1) % len(self.app.sounds)

	def play_sound(self):
		""" Click handler for Play Sound button. """

		sounds = list(self.app.sounds.keys())
		self.app.play_sound(sounds[self.sound_index])

	def draw(self):
		""" Draw method for Sound_Music mode. """

		self.app.draw_text(f'Music: {self.music[self.music_index]}', 10, 10)
		status = 'Playing' if self.app.is_music_playing() else 'Stopped'
		self.app.draw_text(f'Status: {status}', 10, 30)
		sounds = list(self.app.sounds.keys())
		self.app.draw_text(f'Sound: {sounds[self.sound_index]}', 10, 120)

if __name__ == '__main__':
	SoundApp().run()



""" HeadlessWorker worker. """

import time
from jubilee import Worker
from jubilee.misc import Log

class HeadlessWorker(Worker):
	""" HeadlessWorker class for Headless app. """

	def init(self):
		""" HeadlessWorker initializer. """
		
		self.name = f'Worker {int(time.time()) % 100}'
		Log.set_console_level(Log.INFO)

	def process_message(self, message, sender: str=None):
		""" Process a message from app. """

		action = message.get('action', None)
		if action == 'ping':
			Log.info(f'Returned ping {message["id"]}')
			message['action'] = 'pong'
			message['worker'] = self.name
			self.send_message(message)
		else:
			super().process_message(message, sender)



from .app import App
from .worker import Worker
from .mode import Mode
from .base_classes import Sprite, SpritePosition, Animation
from .controls import Control, LabeledControl, Button, HoldButton, CheckButton, ToggleButton, SelectButton
from .log_mode import LogMode
from .shut_down_mode import ShutDownMode
from .misc import Config, Log, Misc, Color



""" Images Jubilee app. """

import random
from jubilee import App, Mode, SpritePosition
from jubilee.base_classes import Sprite

class Bubble:
	""" Bubble class.
			Bubble image source: https://freesvg.org/img/food-blue-bubble.png """

	def __init__(self, app):
		self.app = app
		self.x = random.randint(0, self.app.screen_width)
		self.y = self.app.screen_height + 50

	def process(self):
		""" Process method for Bubble. Moves up the screen until it vanishes off the top. """

		self.x = max(0, min(self.app.screen_width, self.x + random.randint(-1, 1)))
		self.y -= 3
		return (self.y < -10)

	def draw(self):
		""" Draw method for Robot. """

		self.app.blit('bubble', self.x, self.y, position=SpritePosition.Center)

class Robot(Sprite):
	""" Robot class. Subclassed from Sprite for convenience.
			Robot image source: https://commons.wikimedia.org/wiki/File:Koronba_pixel_art.png """

	def __init__(self, app, mode):
		super().__init__(animation=app.get_animation('robot'))
		self.app = app
		self.set_sequence('standing')
		self.x = random.randint(0, app.screen_width)
		self.y = random.randint(0, app.screen_height)
		mode.add_sprite(self)

	def process(self):
		""" Process method for Robot. Moves randomly. """
		
		if random.randint(0, 20) == 0:
			self.x = max(0, min(self.app.screen_width, self.x + random.randint(-5, 5)))
		elif random.randint(0, 20) == 0:
			self.y = max(0, min(self.app.screen_height, self.y + random.randint(-5, 5)))

class Player(Sprite):
	""" Player class. Subclassed from Sprite for convenience. """

	def __init__(self, app, mode):
		super().__init__(animation=app.get_animation('player'))
		self.x = app.screen_center
		self.y = app.screen_middle
		mode.add_sprite(self)

	def process(self):
		""" Process method for Player. Moves and animates based on keyboard input. """
		
		moving = False
		speed = 5
		for direction in ['up', 'down', 'left', 'right']:
			if direction not in self.app.held_keys:
				continue
			moving = True
			self.set_sequence(direction, reset_sequence_frame=False)
			if direction == 'up':
				self.y = max(10, self.y - speed)
			elif direction == 'down':
				self.y = min(self.app.screen_height, self.y + speed)
			elif direction == 'left':
				self.x = max(50, self.x - speed)
			elif direction == 'right':
				self.x = min(self.app.screen_width - 50, self.x + speed)
		if moving is True:
			self.animate()
		else:
			self.set_sequence('standing')

class ImagesApp(App):
	""" Images app. """

	def init(self):
		""" Images app initializer. """
		
		self.add_mode(ImagesMode)

class ImagesMode(Mode):
	""" Images mode. """

	def init(self):
		""" Images mode initializer. """
		
		self.name = 'Images'
		self.bubbles = []
		self.robots = list(Robot(self.app, self) for _ in range(5))			# create 5 robots
		self.player = Player(self.app, self)														# create player

	def process(self):
		""" Process method for Images mode. """

		# process bubbles  (sprites are automatically processed)
		for b in self.bubbles:
			if b.process():
				self.bubbles.remove(b)
		if random.randint(0, 40) == 0:
			self.bubbles.append(Bubble(self.app))

	def draw(self):
		""" Draw method for Images mode. """
		
		self.render_sprites()										# render robots and player
		for b in self.bubbles:									# draw bubbles after (on top of) sprites
			b.draw()

if __name__ == '__main__':
	ImagesApp().run()



""" ImageEffects Jubilee app. """

from jubilee import App, Mode
from jubilee.base_classes import Sprite

class Robot(Sprite):
	""" Robot class. Subclassed from Sprite for convenience.
			Robot image source: https://commons.wikimedia.org/wiki/File:Koronba_pixel_art.png """

	def __init__(self, app, mode):
		super().__init__(animation=app.get_animation('robot'))
		self.app = app
		self.mode = mode
		self.set_sequence('standing')
		self.x = self.app.screen_center
		self.y = self.app.screen_middle + 120
		mode.add_sprite(self)

	def process(self):
		""" Process method for Robot. Moves randomly. """

		graphics_mode = (self.mode.mode_timer // 40) % 4
		graphics_count = self.mode.mode_timer % 40
		self.scale = None
		self.flip_x = False
		self.flip_y = False
		self.rotate = None
		self.hue_shift = None
		if graphics_mode == 0:
			self.scale = 0.5 + (graphics_count + 1) / 20		# scale size from 0.5 to 1.5
		elif graphics_mode == 1:
			self.flip_x = (self.mode.mode_timer % 10 >= 5)
			self.flip_y = (self.mode.mode_timer % 20 >= 10)
		elif graphics_mode == 2:
			self.rotate = 360 * (graphics_count + 1) / 40
		elif graphics_mode == 3:
			self.hue_shift = 360 * (graphics_count + 1) / 40

class ImageEffectsApp(App):
	""" ImageEffects app. """

	def init(self):
		""" Images app initializer. """
		
		self.add_mode(ImageEffectsMode)

class ImageEffectsMode(Mode):
	""" ImageEffects mode. """

	def init(self):
		""" ImageEffects mode initializer. """
		
		self.name = 'ImageEffects'
		self.robot = Robot(self.app, self)

	def draw(self):
		""" Draw method for ImageEffects mode. """

		self.render_sprites()										# render robot

if __name__ == '__main__':
	ImageEffectsApp().run()



""" Headless Jubilee app. """

import random
from jubilee import App, Mode
from jubilee.misc import Log
from headless_worker import HeadlessWorker

class HeadlessApp(App):
	""" Headless app. """

	def init(self):
		Log.set_console_level(Log.INFO)
		self.add_workers([HeadlessWorker, HeadlessWorker])
		self.add_mode(HeadlessMode)
		self.pings = {}

	def process_message(self, message, sender: str=None):
		""" Process message from worker. """

		action = message.get('action')
		if action == 'pong':		# process message
			ping_id = message.get('id')
			if ping_id not in self.pings:
				Log.error(f'Ping ID {ping_id} not pings')
			else:
				self.pings[ping_id] = True
				Log.info(f'Received pong: {ping_id} from {sender}')
		else:
			super().process_message(message, sender)

class HeadlessMode(Mode):
	""" No_Display mode. """

	def init(self):
		self.name = 'Headless'

	def process(self):
		""" Process method for No_Display mode. """
		
		if random.randint(0, 9) > 0:
			return
		try:
			worker = random.choice(list(self.app.workers.keys()))
			ping_id = len(self.app.pings)
			ping_message = {'action': 'ping', 'id': ping_id}
			self.app.send_message(ping_message, worker)
			Log.info(f'Sent ping {ping_id}')
			self.app.pings[ping_id] = False
		except Exception as e:
			Log.error(e)

if __name__ == '__main__':
	HeadlessApp().run()



""" Modes Jubilee app. """

from jubilee import App, Mode
from jubilee.controls import Button

class ModesApp(App):
	""" Modes app. """

	def init(self):
		self.add_modes([MainMode, SubmodeMode])

	def process_message(self, message, sender: str=None):
		""" Process message from worker. """

		action = message.get('action')
		if action == 'custom action':		# process message
			pass
		else:
			super().process_message(message, sender)

class MainMode(Mode):
	""" Main mode. """

	def init(self):
		self.name = 'Main'
		x = self.app.screen_center-50; y = self.app.screen_middle-30; w=100; h=60
		self.add_control(Button('Submode', x, y, w, h, target_mode='Submode'))

	def draw(self):
		""" Draw method for Main mode. """

		self.app.center_text('Main Mode', self.app.screen_middle-50)

class SubmodeMode(Mode):
	""" Submode mode. """

	def init(self):
		self.name = 'Submode'
		x = self.app.screen_center-50; y = self.app.screen_middle-50; w=100; h=60
		self.add_control(Button('Main Mode', x, y, w, h, target_mode='Main'))

	def draw(self):
		""" Draw method for Submode mode. """

		self.app.center_text('Submode', self.app.screen_middle+30)

if __name__ == '__main__':
	ModesApp().run()



""" Submodes Jubilee app. """

import random
from jubilee import App, Mode
from jubilee.misc import Misc

class SubmodesApp(App):
	""" Submodes app. """

	def init(self):
		self.add_mode(MainMode)

	def process_message(self, message, sender: str=None):
		""" Process message from worker. """

		action = message.get('action')
		if action == 'custom action':		# process message
			pass
		else:
			super().process_message(message, sender)

class MainMode(Mode):
	""" Main mode. """

	def init(self):
		self.name = 'Main'
		self.set_submode('generating')
		self.buffer = []
		self.buffer_size = 100

	def enter_generating(self, mode_parameters: dict=None):
		self.buffer = []
	
	def process_generating(self):
		""" Process method for generating submode. """

		color = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
		self.buffer.append(color)
		if len(self.buffer) == self.buffer_size:
			self.set_submode('showing')

	def draw_generating(self):
		""" Draw method for generating submode. """
		
		self.app.center_text(f'Generating: {len(self.buffer)} / 100', 20)
		self.show_buffer()

	def process_showing(self):
		""" Process mode for showing submode. """
		
		rotate = self.buffer.pop()
		self.buffer.insert(0, rotate)

	def click_showing(self, x, y):
		""" Click handler for showing submode. """

		self.set_submode('generating')
	
	def draw_showing(self):
		""" Draw method for showing submode. """

		self.app.center_text(f'Complete (click/touch to restart)', 20)
		self.show_buffer()

	def show_buffer(self):
		""" Draw buffer on display. """
		
		left = self.app.screen_center - self.buffer_size
		self.app.fill_rect(left-2, 48, self.buffer_size*2+4, 54, color='white')
		for i in range(len(self.buffer)):
			color = Misc.get_color(self.buffer[i])
			self.app.fill_rect(left+i*2, 50, 2, 50, color=color)
	
if __name__ == '__main__':
	SubmodesApp().run()



""" Pointer Jubilee app. """

import random
from jubilee import App, Mode

class Circle:
	""" Pointer Circle class. """

	max_radius = 50

	def __init__(self, app, x, y):
		self.app = app
		self.x = x
		self.y = y
		self.radius = 2
		self.color = (random.randint(64, 255), random.randint(64, 255), random.randint(64, 255))

	def process(self) -> bool:
		""" Process function for Circle - expands and incrementally fades colors. """

		self.radius += 1
		return (self.radius == self.max_radius)		# return True if circle should be removed

	def draw(self):
		""" Draw method for Circle - draws filled circle on surface, then blits with alpha. """

		c = self.app.create_surface(self.radius*2, self.radius*2, color=None, alpha_blend=True)
		c.set_alpha(255 * (1.0 - (self.radius / self.max_radius)))
		self.app.fill_circle(self.radius, self.radius, self.radius, color=self.color, dest=c)
		self.app.blit(c, self.x-self.radius, self.y-self.radius)

class PointerApp(App):
	""" Pointer app. """

	def init(self):
		self.add_mode(PointerMode)

class PointerMode(Mode):
	""" Pointer mode. """

	def init(self):
		self.name = 'Pointer'
		self.circles = []

	def click(self, x, y):
		""" Click event handler for Pointer mode. """

		self.circles.append(Circle(self.app, x, y))

	def process(self):
		""" Process method for Pointer mode. """

		for c in self.circles:
			if c.process():
				self.circles.remove(c)

	def draw(self):
		""" Draw method for Pointer mode. """

		for c in self.circles:
			c.draw()
		self.app.center_text('Click anywhere!')

if __name__ == '__main__':
	PointerApp().run()



""" Hello Jubilee app. """

from jubilee import App, Mode
from jubilee.controls import Button
from hello_worker import HelloWorker

class HelloApp(App):
	""" Hello app. """

	def init(self):
		self.add_worker(HelloWorker)
		self.add_mode(HelloMode)
		self.set_mode('Hello')						# note: first mode is selected by default

	def process_message(self, message, sender: str=None):
		""" Process message from worker. """

		action = message.get('action')
		if action == 'custom action':		# process message
			pass
		else:
			super().process_message(message, sender)

class HelloMode(Mode):
	""" Hello mode. """

	def init(self):
		self.name = 'Hello'
		self.click_count = 0
		x = self.app.screen_center - 50; y = self.app.screen_middle - 30
		self.add_control(Button('Hello, World!', x, y, 100, 60, click=self.clicked_hello))

	def clicked_hello(self):
		""" Click handler for Hello button. """
		self.click_count += 1

	def enter(self, mode_parameters: dict=None):
		""" Enter method for Hello mode. """
		super().enter(mode_parameters)

	def set_submode(self, name: str=None, mode_parameters: dict=None):
		""" Set_submode method for Hello mode. """
		super().set_submode(name, mode_parameters)

	def click(self, x: int|float, y: int|float):
		""" Click event handler for Hello mode. """
		pass  # add custom click handling here (controls are handled separately by on_click)

	def process(self):
		""" Process method for Hello mode. """

	def draw(self):
		""" Draw method for Hello mode. """

		# draw text in default font to show click count
		text = f'Click Count: {self.click_count}'
		self.app.center_text(text, self.app.screen_middle + 50)

if __name__ == '__main__':
	HelloApp().run()



""" Controls Jubilee app. """

from jubilee import App, Mode
from jubilee.controls import Button, LabeledControl, HoldButton, CheckButton, ToggleButton, SelectButton

class ControlsApp(App):
	""" Controls app. """

	def init(self):
		self.add_mode(ControlsMode)

class ControlsMode(Mode):
	""" Controls mode. """

	def init(self):
		""" ControlsMode init method. """

		self.click_timer = 0
		self.name = 'Controls'
		self.add_control(Button('Button', 10, 50, 100, 40, click=self.set_click))
		self.add_control(HoldButton('HoldButton', 130, 50, 100, 40, click=self.set_click))
		check_button = CheckButton(10, 100, 20, 20, click=self.set_click)
		self.add_control(LabeledControl('Labeled CheckButton', check_button, offset=150))
		toggle_button = ToggleButton(10, 130, 20, 20, click=self.set_click)
		self.add_control(LabeledControl('Labeled ToggleButton', toggle_button, offset=150))
		select_button = SelectButton(10, 170, 100, 40, ['Option 1', 'Option 2', 'Option 3'], click=self.set_click)
		self.add_control(LabeledControl('Labeled SelectButton', select_button, offset=150))

	def set_click(self):
		""" Starts timer to display "click" message. """

		self.click_timer = 8

	def process(self):
		""" ControlsMode process method. """

		if self.click_timer > 0:
			self.click_timer -= 1

	def draw(self):
		""" ControlsMode draw method. """

		self.app.draw_text('Status:', 100, 20)
		if self.click_timer > 0:
			color = int(255 * self.click_timer / 8)
			self.app.draw_text('Click!', 180, 20, color=(color, color, color))

if __name__ == '__main__':
	ControlsApp().run()



""" Hello worker. """

from jubilee import Worker

class HelloWorker(Worker):
	""" Worker class for Hello app. """

	def init(self):
		""" HelloWorker initializer. """
		
		self.name = 'HelloWorker'

	def process(self):
		""" Regular (high-frequency) worker processing. """
		super().process()

	def process_periodic(self):
		""" Periodic (low-frequency) worker processing. """
		super().process_periodic()

	def process_message(self, message, sender: str=None):
		""" Process a message from app. """

		action = message.get('action', None)
		if action == 'custom action':
			pass
		else:
			super().process_message(message, sender)



""" Script Jubilee app. """

from jubilee import App, Mode
from jubilee.controls import Button

class ScriptApp(App):
	""" Script app. """

	def init(self):
		self.add_modes([MainMode, GraphicMode, SoundMode])
		self.run_script()

class MainMode(Mode):
	""" Main mode. """

	def init(self):
		self.name = 'Main'
		self.set_submode('ask_graphic')

	def enter_ask_graphic(self, mode_parameters: dict=None):
		""" ask_graphic submode enter method. """

		self.remove_controls()
		button_cat = Button('Cat', self.app.screen_center - 150, 50, 100, 50, click=lambda: self.select_graphic('cat'))
		self.add_control(button_cat)
		button_dog = Button('Dog', self.app.screen_center + 50, 50, 100, 50, click=lambda: self.select_graphic('dog'))
		self.add_control(button_dog)
		self.graphic = None
	
	def draw_ask_graphic(self):
		""" ask_graphic submode draw method. """
		
		self.app.center_text('Select Graphic', 20)

	def select_graphic(self, graphic):
		""" Select graphic method. """
		
		self.app.advance_scene(1 if graphic == 'cat' else 2)

	def enter_ask_sound(self, mode_parameters: dict=None):
		""" ask_sound submode enter method. """

		self.remove_controls()
		button_meow = Button('Meow', self.app.screen_center - 150, 50, 100, 50, click=lambda: self.select_sound('meow'))
		self.add_control(button_meow)
		button_woof = Button('Woof', self.app.screen_center + 50, 50, 100, 50, click=lambda: self.select_sound('woof'))
		self.add_control(button_woof)
		self.graphic = None

	def draw_ask_sound(self):
		""" ask_sound submode draw method. """
		
		self.app.center_text('Select Sound', 20)

	def select_sound(self, sound):
		self.app.advance_scene(1 if sound == 'meow' else 2)

	def draw_finish(self):
		""" finish submode draw method. """
		
		self.app.center_text('The End', 20)

	def enter_finish(self, mode_parameters: dict=None):
		""" finish submode enter method. """

		self.remove_controls()
		button_restart = Button('Restart', self.app.screen_center - 50, 50, 100, 50, click=self.restart_mode)
		self.add_control(button_restart)

	def restart_mode(self):
		""" Restart mode. """

		self.app.select_scene('Select_Graphic')

class GraphicMode(Mode):
	""" Graphic mode. """

	def init(self):
		self.name = 'Graphic'

	def enter(self, mode_parameters: dict=None):
		""" Mode graphic enter method. """

		self.graphic = mode_parameters.get('graphic')

	def click(self, x, y):
		""" Mode graphic click handler. """
		
		self.app.select_scene('Select_Sound')

	def draw(self):
		""" Mode graphic draw method. """
		
		if self.graphic == 'dog':
			self.app.blit('dog', 0, 0)
		elif self.graphic == 'cat':
			self.app.blit('cat', 0, 0)

class SoundMode(Mode):
	""" Sound mode. """

	def init(self):
		self.name = 'Sound'

	def enter(self, mode_parameters: dict=None):
		""" Mode sound enter method. """

		self.sound = mode_parameters.get('sound')
		self.remove_controls()
		button_play = Button('Play Sound', self.app.screen_center - 50, 50, 100, 50, click=self.play_sound)
		self.add_control(button_play)
		button_continue = Button('Continue', self.app.screen_center - 50, 120, 100, 50, click=lambda: self.app.select_scene('Finish'))
		self.add_control(button_continue)

	def play_sound(self):
		""" Play sound method. """

		self.app.play_sound(self.sound)
			
	def draw(self):
		""" Mode sound draw method. """

		self.app.center_text(f'Ready to play sound: {self.sound}', 20)

if __name__ == '__main__':
	ScriptApp().run()



""" Screen_Rotation Jubilee app. """

from jubilee import App, Mode

class ScreenRotationApp(App):
	""" ScreenRotation app. """

	def init(self):
		self.add_mode(ScreenRotationMode)

class ScreenRotationMode(Mode):
	""" ScreenRotation mode. """

	def init(self):
		self.name = 'Screen Rotation'

	def draw(self):
		""" Draw method for ScreenRotation mode. """

		self.app.center_text('Rotation', self.app.screen_middle - 20)

if __name__ == '__main__':
	ScreenRotationApp().run()



""" Jubilee Worker class. """

import datetime, json, multiprocessing, os, platform, queue, subprocess, sys, time
import __main__
from .misc import Config, Log

class Worker:
	""" Worker class. """

	config_defaults = {'headless': False,
		'nosound': False,
		'screen_resolution': [320, 240],
		'screen_scale': [[0, 319, -1], [0, 239, 1]],
		'screen_rotation': 0,
		'app_process_fps': 20,
		'app_draw_fps': 20, 'modal': True,
		'worker_process_fps': 20,
		'worker_process_periodic_fps': 1,
		'persist_app_state': True,
		'app_state_filename': 'app_state.json',
		'app_state_start_filename': 'app_state_start.json',
		'keyboard_input': True,
		'log_rotation': 'daily',
		'font_size': 14,
		'font_size_desktop': 15,
		'wifi_watchdog': False,
		'wifi_ping_target': None,
		'wifi_ping_interval': 300,
		'wifi_interface': 'wlan0',
		'wifi_ping_count': 3,
		'wifi_ping_timeout': 5,
		'wifi_reboot_enabled': True,
		'wifi_reboot_daily_max': 3,
		'wifi_reboot_after_failures': 3
	}

	def __init__(self, app_queue, worker_queue, config_manager=False, log_manager=False, wifi_manager=False):
		self.name = 'Worker'
		self.app_queue = app_queue
		self.worker_queue = worker_queue
		self.config_manager = config_manager
		self.log_manager = log_manager
		self.wifi_manager = wifi_manager
		self.worker_process = None
		self.base_path = os.path.dirname(os.path.realpath(__main__.__file__))
		self.project_path = Config.project_path or self.base_path
		self.config_filename = os.path.join(self.project_path, 'config.toml')
		self.config = Config.load(self.config_filename, defaults=self.config_defaults)
		self.config_date = None
		self.log_date = None
		if os.path.isfile(self.config_filename) is True:
			self.config_date = os.path.getmtime(self.config_filename)
		self.last_periodic = None
		self.wifi_last_check = None
		self.wifi_consecutive_failures = 0
		self.wifi_reboot_count_today = 0
		self.wifi_reboot_date = None
		self.wifi_recovering = False
		self.wifi_connected = None
		self.wifi_boot_grace = True
		self.init()

		# start process
		self.worker_process = multiprocessing.Process(target=self.run)
		self.worker_process.daemon = True
		self.worker_process.start()

	def init(self):
		""" Worker-specific initializer that runs before the worker starts.
					State changes made here are visible to the App.
		"""

	def start_worker(self):
		""" Worker-specific initializer that runs at the start of run loop.
					State changes made here are *not* visible to the App.
		"""

	def run(self):
		""" Worker run loop. """

		try:
			# restore class-level paths for child process
			Log.project_path = self.project_path
			Config.project_path = self.project_path

			Log.info('Starting')
			self.start_worker()

			while True:
				loop_start = time.time()
				self.receive_messages()

				# call main process function
				self.on_process()

				# call periodic process function occasionally
				process_periodic_fps = self.config.get('worker_process_periodic_fps', 1)
				if process_periodic_fps is not None:
					elapsed = time.time() - (self.last_periodic or 0)
					if elapsed >= 1 / process_periodic_fps:
						self.on_process_periodic()

				# delay to next loop
				loop_time = time.time() - loop_start
				delay = 1 / max(1, int(self.config.get('worker_process_fps', 20))) - loop_time
				if delay > 0:
					time.sleep(delay)

		except Exception as e:
			Log.error(e)

	def manage_config(self):
		""" Config manager. """

		if self.config_manager is False:
			return
		if os.path.isfile(self.config_filename) is False:
			self.config_date = None
			return
		config_date = os.path.getmtime(self.config_filename)
		if config_date != self.config_date:
			Log.info(f'Loading config ({self.config_date} != {config_date})')
			self.config_date = config_date
			self.config = Config.load(self.config_filename, defaults=self.config_defaults)
			self.send_updated_config()

	def manage_log(self):
		""" Log manager. """

		if self.log_manager is False:
			return
		if os.path.isfile(Log.get_filename()) is False:
			self.log_date = None
			return
		self.log_date = self.log_date or datetime.datetime.now()
		rotation = self.config.get('log_rotation', 'daily')
		rotate = False
		log_format = None
		if rotation == 'daily':
			rotate = (datetime.datetime.now().strftime('%Y%m%d') != self.log_date.strftime('%Y%m%d'))
			log_format = '%Y%m%d'
		elif rotation == 'monthly':
			rotate = (datetime.datetime.now().strftime('%Y%m') != self.log_date.strftime('%Y%m'))
			log_format = '%Y%m'
		elif rotation == 'hourly':
			rotate = (datetime.datetime.now().strftime('%Y%m%d%H') != self.log_date.strftime('%Y%m%d%H'))
			log_format = '%Y%m%d%H'
		else:
			Log.warning(f'Unrecognized log_rotation value: {rotation}')
			return
		if rotate is True:
			filename = f'log_{datetime.datetime.now().strftime(log_format)}.txt'
			Log.backup(backup_filename=filename)
			self.log_date = datetime.datetime.now()

	def manage_wifi(self):
		""" WiFi watchdog manager. Periodically checks connectivity and performs
				escalating recovery when WiFi is lost. Also serves as keepalive traffic
				to prevent the WiFi firmware from entering problematic idle states.
				Only operates on Linux. """

		if self.wifi_manager is False:
			return
		if platform.system() != 'Linux':
			return
		if self.config.get('wifi_watchdog', False) is not True:
			return

		# timer-based throttling: only run at configured interval
		now = time.time()
		interval = self.config.get('wifi_ping_interval', 300)
		if self.wifi_last_check is not None and (now - self.wifi_last_check) < interval:
			return
		self.wifi_last_check = now

		# boot grace: skip first check to allow network to come up
		if self.wifi_boot_grace is True:
			self.wifi_boot_grace = False
			Log.info('WiFi watchdog: boot grace period, skipping first check')
			return

		# guard against re-entry during recovery
		if self.wifi_recovering is True:
			return

		# resolve ping target
		target = self.config.get('wifi_ping_target')
		if target is None:
			target = self._wifi_detect_gateway()
		if target is None:
			Log.warning('WiFi watchdog: no ping target configured and could not detect gateway')
			return

		interface = self.config.get('wifi_interface', 'wlan0')
		ping_count = self.config.get('wifi_ping_count', 3)
		ping_timeout = self.config.get('wifi_ping_timeout', 5)

		# connectivity check (also serves as keepalive traffic)
		connected = self._wifi_ping(target, interface, ping_count, ping_timeout)

		if connected:
			if self.wifi_consecutive_failures > 0:
				Log.info(f'WiFi watchdog: connectivity restored after {self.wifi_consecutive_failures} failure(s)')
			self.wifi_consecutive_failures = 0
			if self.wifi_connected is not True:
				self.wifi_connected = True
				self._wifi_send_status(True, target)
			return

		# connectivity failed
		self.wifi_consecutive_failures += 1
		self.wifi_connected = False
		Log.warning(f'WiFi watchdog: connectivity check failed (consecutive: {self.wifi_consecutive_failures})')
		self._wifi_send_status(False, target)

		# begin escalating recovery
		self.wifi_recovering = True
		try:
			self._wifi_escalating_recovery(target, interface, ping_count, ping_timeout)
		finally:
			self.wifi_recovering = False

	def _wifi_detect_gateway(self):
		""" Detects default gateway IP address via ip route. """

		try:
			result = subprocess.run(
				['ip', 'route', 'show', 'default'],
				capture_output=True, text=True, timeout=5
			)
			if result.returncode == 0 and result.stdout.strip():
				parts = result.stdout.strip().split()
				if len(parts) >= 3 and parts[0] == 'default' and parts[1] == 'via':
					return parts[2]
		except Exception as e:
			Log.debug(f'WiFi watchdog: gateway detection failed: {e}')
		return None

	def _wifi_ping(self, target, interface, count, timeout):
		""" Pings target through interface. Returns True if reachable. """

		try:
			result = subprocess.run(
				['ping', '-c', str(count), '-W', str(timeout), '-I', interface, target],
				capture_output=True, timeout=count * timeout + 10
			)
			return result.returncode == 0
		except subprocess.TimeoutExpired:
			Log.warning('WiFi watchdog: ping subprocess timed out')
			return False
		except Exception as e:
			Log.error(f'WiFi watchdog: ping failed: {e}')
			return False

	def _wifi_send_status(self, connected, target):
		""" Sends WiFi status message to App. """

		self.send_message({
			'action': 'wifi status',
			'connected': connected,
			'target': target,
			'consecutive_failures': self.wifi_consecutive_failures
		})

	def _wifi_send_recovery(self, step):
		""" Sends WiFi recovery action message to App. """

		self.send_message({
			'action': 'wifi recovery',
			'step': step,
			'consecutive_failures': self.wifi_consecutive_failures
		})

	def _wifi_run_command(self, args, timeout=30, description='command'):
		""" Runs a subprocess command with timeout. Returns True on success. """

		try:
			result = subprocess.run(args, capture_output=True, text=True, timeout=timeout)
			if result.returncode != 0:
				Log.debug(f'WiFi watchdog: {description} returned {result.returncode}: {result.stderr.strip()}')
			return result.returncode == 0
		except subprocess.TimeoutExpired:
			Log.warning(f'WiFi watchdog: {description} timed out after {timeout}s')
			return False
		except Exception as e:
			Log.error(f'WiFi watchdog: {description} failed: {e}')
			return False

	def _wifi_interruptible_sleep(self, seconds):
		""" Sleeps for the specified duration while remaining responsive to exit messages. """

		end = time.time() + seconds
		while time.time() < end:
			try:
				message = self.app_queue.get_nowait()
				parsed = json.loads(message)
				if parsed.get('action') == 'exit':
					self.exit()
			except queue.Empty:
				pass
			remaining = end - time.time()
			if remaining > 0:
				time.sleep(min(remaining, 0.5))

	def _wifi_escalating_recovery(self, target, interface, ping_count, ping_timeout):
		""" Performs escalating WiFi recovery steps, remaining responsive to exit messages. """

		# step 1: nmcli device connect
		Log.info('WiFi watchdog: Step 1 - nmcli reconnect')
		self._wifi_run_command(
			['nmcli', 'device', 'connect', interface],
			timeout=30, description='nmcli reconnect'
		)
		self._wifi_send_recovery('nmcli reconnect')
		self._wifi_interruptible_sleep(15)
		if self._wifi_ping(target, interface, ping_count, ping_timeout):
			Log.info('WiFi watchdog: recovered via nmcli reconnect')
			self.wifi_consecutive_failures = 0
			self.wifi_connected = True
			self._wifi_send_status(True, target)
			return

		# step 2: interface restart (ip link down/up)
		Log.info('WiFi watchdog: Step 2 - interface restart')
		self._wifi_run_command(
			['ip', 'link', 'set', interface, 'down'],
			timeout=10, description='interface down'
		)
		self._wifi_interruptible_sleep(2)
		self._wifi_run_command(
			['ip', 'link', 'set', interface, 'up'],
			timeout=10, description='interface up'
		)
		self._wifi_send_recovery('interface restart')
		self._wifi_interruptible_sleep(15)
		if self._wifi_ping(target, interface, ping_count, ping_timeout):
			Log.info('WiFi watchdog: recovered via interface restart')
			self.wifi_consecutive_failures = 0
			self.wifi_connected = True
			self._wifi_send_status(True, target)
			return

		# step 3: restart NetworkManager
		Log.info('WiFi watchdog: Step 3 - NetworkManager restart')
		self._wifi_run_command(
			['systemctl', 'restart', 'NetworkManager'],
			timeout=30, description='NetworkManager restart'
		)
		self._wifi_send_recovery('NetworkManager restart')
		self._wifi_interruptible_sleep(20)
		if self._wifi_ping(target, interface, ping_count, ping_timeout):
			Log.info('WiFi watchdog: recovered via NetworkManager restart')
			self.wifi_consecutive_failures = 0
			self.wifi_connected = True
			self._wifi_send_status(True, target)
			return

		# step 4: reload WiFi driver (modprobe -r / modprobe brcmfmac)
		Log.info('WiFi watchdog: Step 4 - driver reload')
		self._wifi_run_command(
			['modprobe', '-r', 'brcmfmac'],
			timeout=10, description='modprobe remove'
		)
		self._wifi_interruptible_sleep(3)
		self._wifi_run_command(
			['modprobe', 'brcmfmac'],
			timeout=10, description='modprobe load'
		)
		self._wifi_interruptible_sleep(10)
		self._wifi_run_command(
			['systemctl', 'restart', 'NetworkManager'],
			timeout=30, description='NetworkManager restart after driver reload'
		)
		self._wifi_send_recovery('driver reload')
		self._wifi_interruptible_sleep(20)
		if self._wifi_ping(target, interface, ping_count, ping_timeout):
			Log.info('WiFi watchdog: recovered via driver reload')
			self.wifi_consecutive_failures = 0
			self.wifi_connected = True
			self._wifi_send_status(True, target)
			return

		# step 5: reboot (conditional)
		Log.error('WiFi watchdog: all recovery steps failed')
		reboot_enabled = self.config.get('wifi_reboot_enabled', True)
		reboot_daily_max = self.config.get('wifi_reboot_daily_max', 3)
		reboot_after_failures = self.config.get('wifi_reboot_after_failures', 3)

		if reboot_enabled is not True:
			Log.warning('WiFi watchdog: reboot disabled in config')
			return

		if self.wifi_consecutive_failures < reboot_after_failures:
			Log.info(f'WiFi watchdog: deferring reboot ({self.wifi_consecutive_failures}/{reboot_after_failures} consecutive failures)')
			return

		# check daily reboot cap
		today = time.strftime('%Y%m%d')
		if self.wifi_reboot_date != today:
			self.wifi_reboot_date = today
			self.wifi_reboot_count_today = 0

		if self.wifi_reboot_count_today >= reboot_daily_max:
			Log.error(f'WiFi watchdog: daily reboot limit ({reboot_daily_max}) reached, manual intervention required')
			return

		self.wifi_reboot_count_today += 1
		Log.warning(f'WiFi watchdog: initiating reboot ({self.wifi_reboot_count_today}/{reboot_daily_max} today)')
		self._wifi_send_recovery('reboot')
		self._wifi_interruptible_sleep(1)
		self._wifi_run_command(
			['sudo', 'shutdown', '-r', 'now'],
			timeout=5, description='reboot'
		)

	def on_process(self):
		""" Regular (high-frequency) worker processing event receiver.
				This merely wraps process() but is a placeholder for future
				functionality and for consistency with other event receivers. """

		self.process()

	def process(self):
		""" Regular (high-frequency) worker processing stub function. """

	def on_process_periodic(self):
		""" Periodic (low-frequency) worker processing event receiver. """

		self.last_periodic = time.time()
		if self.config_manager is True:
			self.manage_config()
		if self.log_manager is True:
			self.manage_log()
		if self.wifi_manager is True:
			self.manage_wifi()
		self.process_periodic()

	def process_periodic(self):
		""" Periodic (low-frequency) worker processing stub function. """

	def exit(self, code=0):
		sys.exit(code)

	# messaging with app

	def send_message(self, message: str|dict):
		""" Send a message to the app. """

		if isinstance(message, str):
			message = {'action': message}
		try:
			self.worker_queue.put(json.dumps(message, ensure_ascii=True))
		except Exception as e:
			Log.error(f'Failed to send message: {e}')

	def receive_messages(self):
		""" Receive messages from app. """

		while True:
			try:
				message = self.app_queue.get_nowait()
				self.process_message(json.loads(message), sender='App')
			except queue.Empty:
				return
			except Exception as e:
				Log.error(e)
				continue

	def process_message(self, message: dict, sender: str=None):
		""" Process a message from app. This method can be extended in subclass. """

		action = message.get('action')
		if action == 'update config':
			key = message.get('key')
			value = message.get('value')
			self.update_config(key, value)
		elif action == 'config updated':
			self.config = message.get('config', {})
		elif action == 'exit':
			self.exit()
		else:
			Log.warning(f'Received unknown message: {message}')

	def update_config(self, key, value):
		self.config[key] = value
		self.write_config()

	def write_config(self):
		Config.save(self.config, self.config_filename)
		self.config_date = os.path.getmtime(self.config_filename) if os.path.isfile(self.config_filename) else None
		self.send_updated_config()

	def send_updated_config(self):
		message = {'action': 'config updated', 'config': self.config}
		self.send_message(message)
